/*
 * FILE:
 *   CoercriGfx.cs
 *
 * PURPOSE:
 *   Class representing Coercri graphics system
 *
 * AUTHOR:
 *   Stephen Thompson <stephen@solarflare.org.uk>
 *
 * CREATED:
 *   18-Jun-2011
 *
 * COPYRIGHT:
 *   Copyright (C) Stephen Thompson, 2011.
 *
 *   This file is part of Mips2cs. Mips2cs is distributed under the terms
 *   of the Boost Software License, Version 1.0, the text of which
 *   appears below.
 *
 *   Boost Software License - Version 1.0 - August 17th, 2003
 *   
 *   Permission is hereby granted, free of charge, to any person or organization
 *   obtaining a copy of the software and accompanying documentation covered by
 *   this license (the "Software") to use, reproduce, display, distribute,
 *   execute, and transmit the Software, and to prepare derivative works of the
 *   Software, and to permit third-parties to whom the Software is furnished to
 *   do so, all subject to the following:
 *
 *   The copyright notices in the Software and this entire statement, including
 *   the above license grant, this restriction and the following disclaimer,
 *   must be included in all copies of the Software, in whole or in part, and
 *   all derivative works of the Software, unless such copies or derivative
 *   works are solely in the form of machine-executable object code generated by
 *   a source language processor.
 *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 *   SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 *   FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *   DEALINGS IN THE SOFTWARE.
 *
 */

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Collections.Generic;

class CoercriGfx
{
    class MyGraphic
    {
        public Texture2D tex;
        public int hx;
        public int hy;
    }

    GraphicsDevice gfx_dev;
    GraphicsDeviceManager gfx_dev_manager;
    GraphicsAdapter gfx_adapter;
    GameWindow window;
    
    Viewport default_viewport;
      
    SpriteBatch sprite_batch;
    bool sprite_begun;
    
    Dictionary<int, MyGraphic> graphics;
    int next_graphic_id;
    
    // used for line and rectangle drawing:
    const int VertBufferSize = 500;
    VertexPositionColor[] vertices = new VertexPositionColor[VertBufferSize];
    int vertBufferPos = 0;
    BasicEffect basicEffect;
    PrimitiveType prim_type;
    bool prim_begun;
    
    public CoercriGfx(GraphicsDevice dev, GraphicsDeviceManager mgr, GraphicsAdapter adapt, GameWindow win)
    {
        gfx_dev = dev;
        gfx_dev_manager = mgr;
        gfx_adapter = adapt;
        window = win;
        
        default_viewport = gfx_dev.Viewport;
        
        prim_begun = false;
        prim_type = PrimitiveType.LineList;
        basicEffect = new BasicEffect(gfx_dev);
        basicEffect.VertexColorEnabled = true;
        basicEffect.Projection = Matrix.CreateOrthographicOffCenter(0,
            gfx_dev.Viewport.Width,
            gfx_dev.Viewport.Height,
            0,
            0,
            1);
        
        sprite_batch = new SpriteBatch(gfx_dev);
        sprite_begun = false;
        
        graphics = new Dictionary<int, MyGraphic>();
        next_graphic_id = 2;

        // create a 1x1 graphic for plotting single points
        Texture2D tex = new Texture2D(gfx_dev, 1, 1, false, SurfaceFormat.Color);
        Color[] cols = new Color[1];
        cols[0] = new Color(255,255,255,255);
        tex.SetData(cols);
        MyGraphic g = new MyGraphic();
        g.tex = tex;
        g.hx = 0;
        g.hy = 0;
        graphics[1] = g;
    }
    
    public int GetNumFullScreenModes()
    {
        int i = 0;
        foreach (DisplayMode dm in gfx_adapter.SupportedDisplayModes)
        {
            i++;
        }
        return i;
    }
    
    public void GetFullScreenModes(MipsVM mips, int out_addr)
    {
        foreach (DisplayMode dm in gfx_adapter.SupportedDisplayModes)
        {
            mips.WriteWord(out_addr, dm.Width);
            out_addr += 4;
            mips.WriteWord(out_addr, dm.Height);
            out_addr += 4;
        }
    }
        
    public void GetDesktopMode(MipsVM mips, int out_addr)
    {
        mips.WriteWord(out_addr, gfx_adapter.CurrentDisplayMode.Width);
        out_addr += 4;
        mips.WriteWord(out_addr, gfx_adapter.CurrentDisplayMode.Height);
    }
        
    public void CreateWindow(MipsVM mips, int width, int height, bool resizable, bool fullscreen, int title_addr)
    {
        gfx_dev_manager.PreferredBackBufferWidth = width;
        gfx_dev_manager.PreferredBackBufferHeight = height;
        gfx_dev_manager.IsFullScreen = fullscreen;

        window.AllowUserResizing = resizable;            
            
        // title
        string title = "";
        while (true)
        {
            int ch = mips.ReadByteU(title_addr++);
            if (ch == 0) break;
            title += (char)ch;
        }
        window.Title = title;

        gfx_dev_manager.ApplyChanges();

        basicEffect.Projection = Matrix.CreateOrthographicOffCenter(0,
            gfx_dev.Viewport.Width,
            gfx_dev.Viewport.Height,
            0,
            0,
            1);
    }

    public void SwitchWindow(int width, int height, bool fullscreen)
    {
        gfx_dev_manager.PreferredBackBufferWidth = width;
        gfx_dev_manager.PreferredBackBufferHeight = height;
        gfx_dev_manager.IsFullScreen = fullscreen;
        gfx_dev_manager.ApplyChanges();

        basicEffect.Projection = Matrix.CreateOrthographicOffCenter(0,
            gfx_dev.Viewport.Width,
            gfx_dev.Viewport.Height,
            0,
            0,
            1);
    }

    public bool IsFullScreen()
    {
        return gfx_dev_manager.IsFullScreen;
    }
        
    public int CreateGraphic(MipsVM mips, int pixel_addr, int width, int height, int hx, int hy)
    {
        // Returns an ID number
        Texture2D tex = new Texture2D(gfx_dev, width, height, false, SurfaceFormat.Color);
        
        // load gfx data
        Color[] cols = new Color[width * height];
        for (int i = 0; i < width * height; ++i)
        {
            cols[i] = new Color(mips.ReadByteU(pixel_addr + i*4),
                                mips.ReadByteU(pixel_addr + i*4 + 1),
                                mips.ReadByteU(pixel_addr + i*4 + 2),
                                mips.ReadByteU(pixel_addr + i*4 + 3));
        }
        tex.SetData(cols);

        // create the graphic.
        MyGraphic g = new MyGraphic();
        g.tex = tex;
        g.hx = hx;
        g.hy = hy;
        int id = next_graphic_id++;
        graphics[id] = g;
        return id;
    }

    public void DestroyGraphic(int id)
    {
        graphics[id].tex.Dispose();
        graphics.Remove(id);
    }
    
    public void GetWindowSize(MipsVM mips, int w_addr, int h_addr)
    {
        mips.WriteWord(w_addr, gfx_dev.PresentationParameters.BackBufferWidth);
        mips.WriteWord(h_addr, gfx_dev.PresentationParameters.BackBufferHeight);
    }

    public void SetClipRectangle(int x, int y, int w, int h)
    {
        Viewport v = new Viewport();
        v.X = x;
        v.Y = y;
        v.Width = w;
        v.Height = h;
        gfx_dev.Viewport = v;
    }

    public void ClearClipRectangle()
    {
        gfx_dev.Viewport = default_viewport;
    }
    
    public void GetClipRectangle(ref int x, ref int y, ref int w, ref int h)
    {
        x = gfx_dev.Viewport.X;
        y = gfx_dev.Viewport.Y;
        w = gfx_dev.Viewport.Width;
        h = gfx_dev.Viewport.Height;
    }
    
    public void ClearScreen(Color col)
    {
        gfx_dev.Clear(col);
    }
    
    private void SetupPrim(PrimitiveType pt)
    {
        if (sprite_begun)
        {
            sprite_batch.End();
            sprite_begun = false;
        }
            
        if (prim_begun)
        {
            if (prim_type == pt)
            {
                return;
            }
            FlushPrim();
        }
        
        prim_begun = true;
        prim_type = pt;

        basicEffect.CurrentTechnique.Passes[0].Apply();
    }

    private void FlushPrim()
    {
        if (!prim_begun) return;
        if (vertBufferPos == 0) return;

        int Count = vertBufferPos;
        if (prim_type == PrimitiveType.LineList) Count /= 2;
        else Count /= 3;

        gfx_dev.DrawUserPrimitives<VertexPositionColor>(prim_type, vertices, 0, Count);
        vertBufferPos = 0;
    }
    
    private void SetupSprite()
    {
        if (prim_begun)
        {
            FlushPrim();
            prim_begun = false;
        }
        
        if (!sprite_begun)
        {
            sprite_batch.Begin(SpriteSortMode.Deferred, BlendState.NonPremultiplied);
            sprite_begun = true;
        }
    }

    public void EndBatches()
    {
        if (prim_begun)
        {
            FlushPrim();
            prim_begun = false;
        }

        if (sprite_begun)
        {
            sprite_batch.End();
            sprite_begun = false;
        }
    }
        
    public void PlotPixel(int x, int y, Color col)
    {
        SetupSprite();
        Texture2D tex = graphics[1].tex;
        sprite_batch.Draw(tex, new Vector2(x, y), col);
    }
    
    public void DrawLine(int x1, int y1, int x2, int y2, Color col)
    {
        SetupPrim(PrimitiveType.LineList);
        if (vertBufferPos > VertBufferSize - 2) FlushPrim();

        vertices[vertBufferPos].Position = new Vector3(x1, y1, 0);
        vertices[vertBufferPos].Color = col;
        vertices[vertBufferPos + 1].Position = new Vector3(x2, y2, 0);
        vertices[vertBufferPos + 1].Color = col;

        vertBufferPos += 2;
    }

    public void FillRectangle(int x1, int y1, int x2, int y2, Color col)
    {
        SetupPrim(PrimitiveType.TriangleList);
        if (vertBufferPos > VertBufferSize - 6) FlushPrim();

        Vector3 tl = new Vector3(x1, y1, 0);
        Vector3 tr = new Vector3(x2, y1, 0);
        Vector3 bl = new Vector3(x1, y2, 0);
        Vector3 br = new Vector3(x2, y2, 0);

        vertices[vertBufferPos].Position = tr;
        vertices[vertBufferPos].Color = col;
        vertices[vertBufferPos + 1].Position = bl;
        vertices[vertBufferPos + 1].Color = col;
        vertices[vertBufferPos + 2].Position = tl;
        vertices[vertBufferPos + 2].Color = col;

        vertices[vertBufferPos + 3].Position = tr;
        vertices[vertBufferPos + 3].Color = col;
        vertices[vertBufferPos + 4].Position = br;
        vertices[vertBufferPos + 4].Color = col;
        vertices[vertBufferPos + 5].Position = bl;
        vertices[vertBufferPos + 5].Color = col;

        vertBufferPos += 6;
    }

    public void DrawGraphic(int x, int y, int id)
    {
        SetupSprite();
        MyGraphic g = graphics[id];
        sprite_batch.Draw(g.tex, new Vector2(x - g.hx, y - g.hy), Color.White);
    }

    public int GetGraphicWidth(int id)
    {
        return graphics[id].tex.Width;
    }

    public int GetGraphicHeight(int id)
    {
        return graphics[id].tex.Height;
    }

    public void GetGraphicHandle(int id, ref int x, ref int y)
    {
        MyGraphic g = graphics[id];
        x = g.hx;
        y = g.hy;
    }

    public void GetGraphicPixels(MipsVM mips, int id, int addr)
    {
        MyGraphic g = graphics[id];
        int width = g.tex.Width;
        int height = g.tex.Height;

        Color[] cols = new Color[width * height];
        g.tex.GetData(cols);

        for (int i = 0; i < width * height; ++i)
        {
            mips.WriteByte(addr + i*4, cols[i].R);
            mips.WriteByte(addr + i*4 + 1, cols[i].G);
            mips.WriteByte(addr + i*4 + 2, cols[i].B);
            mips.WriteByte(addr + i*4 + 3, cols[i].A);
        }
    }
    
    // TODO: Fonts/Text.

}
