#!/usr/bin/env runhaskell

-- FILE:
--   runtests.hs
--
-- PURPOSE:
--   Script to run the Mips2cs test harness and compare results
--
-- AUTHOR:
--   Stephen Thompson <stephen@solarflare.org.uk>
--
-- CREATED:
--   10-Jun-2011
--
-- COPYRIGHT:
--   Copyright (C) Stephen Thompson, 2011.
--
--   This file is part of Mips2cs. Mips2cs is distributed under the terms
--   of the Boost Software License, Version 1.0, the text of which
--   appears below.
--
--   Boost Software License - Version 1.0 - August 17th, 2003
--   
--   Permission is hereby granted, free of charge, to any person or organization
--   obtaining a copy of the software and accompanying documentation covered by
--   this license (the "Software") to use, reproduce, display, distribute,
--   execute, and transmit the Software, and to prepare derivative works of the
--   Software, and to permit third-parties to whom the Software is furnished to
--   do so, all subject to the following:
--
--   The copyright notices in the Software and this entire statement, including
--   the above license grant, this restriction and the following disclaimer,
--   must be included in all copies of the Software, in whole or in part, and
--   all derivative works of the Software, unless such copies or derivative
--   works are solely in the form of machine-executable object code generated by
--   a source language processor.
--
--   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
--   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
--   FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
--   SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
--   FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
--   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
--   DEALINGS IN THE SOFTWARE.


import Data.List
import System.Directory
import System.Environment
import System.Exit
import System.Process

mySystem :: String -> IO ExitCode
mySystem cmd = putStrLn cmd >> system cmd

deleteOutputFiles :: String -> IO ()
deleteOutputFiles testName = do
  mySystem $ "rm -f output/" ++ testName ++ ".*"
  return ()

compileMipsC :: String -> IO ()
compileMipsC testName = do
  mySystem $ "mips-elf-gcc input/" ++ testName
             ++ " -Wl,--start-group ../console_lib/lib/console_lib.a -lc -Wl,--end-group"
             ++ " -o output/" ++ testName ++ ".mips"
             ++ " -Wl,--emit-relocs"
             ++ " -msoft-float"
             ++ " -O2"
  return ()

compileMipsCXX :: String -> IO ()
compileMipsCXX testName = do
  mySystem $ "mips-elf-g++ input/" ++ testName
               ++ " -Wl,--start-group -lstdc++ -lc ../console_lib/lib/console_lib.a -Wl,--end-group"
               ++ " -o output/" ++ testName ++ ".mips"
               ++ " -Wl,--emit-relocs"
               ++ " -msoft-float"
               ++ " -O2"
  return ()

compileMips :: String -> IO ()
compileMips testName = do
    if ".c" `isSuffixOf` testName || ".s" `isSuffixOf` testName then
        compileMipsC testName
     else if ".cpp" `isSuffixOf` testName then
              compileMipsCXX testName
           else do
             putStrLn $ "*** ERROR: Unknown file type: " ++ testName
             exitFailure

runMips2cs :: String -> String -> IO ()
runMips2cs testName optLevel = do
  mySystem $ "../src/Main -O" ++ optLevel 
             ++ " output/" ++ testName ++ ".mips"
             ++ " output/" ++ testName ++ ".cs"
  return ()

compileCs :: String -> IO ()
compileCs testName = do
  mySystem $ "gmcs output/" ++ testName ++ ".cs"
             ++ " ../console_lib/cs/ConsoleLibCallbacks.cs"
             ++ " ../console_lib/cs/ConsoleLibMain.cs"
             ++ " -out:output/" ++ testName ++ ".exe"
             ++ " -warn:0"
  return ()

runCompiledProgram :: String -> IO ()
runCompiledProgram testName = do
  result <- mySystem $ "output/" ++ testName ++ ".exe >output/" ++ testName ++ ".txt"
  case result of 
    ExitSuccess -> return ()
    ExitFailure _ -> do
                putStrLn $ "*** PROBLEM: " ++ testName ++ ".exe failed to run."
                exitFailure

compareOutput :: String -> IO ()
compareOutput testName = do
  result <- mySystem $ "diff -u reference/" ++ testName ++ ".txt output/" ++ testName ++ ".txt"
  case result of 
    ExitSuccess -> return ()
    ExitFailure 1 -> putStrLn $ "*** PROBLEM: " ++ testName ++ ".txt is different."
    ExitFailure _ -> putStrLn $ "*** PROBLEM: Failed to compare " ++ testName ++ ".txt."
  case result of 
    ExitSuccess -> return ()
    ExitFailure _ -> exitFailure

runTest :: String -> String -> IO ()
runTest optLevel testName = do
  putStrLn ""
  putStrLn $ "Running: " ++ testName
  deleteOutputFiles testName
  compileMips testName
  runMips2cs testName optLevel
  compileCs testName
  runCompiledProgram testName
  compareOutput testName

printUsage :: IO a
printUsage = do
  putStrLn "Usage: runtests.hs [-O<n>] [filenames]"
  exitFailure

-- rudimentary argument parsing.
processArgs :: IO (String, [String])
processArgs = do
  xs <- getArgs
  let (opts, files) = partition ("-" `isPrefixOf`) xs
  optLevel <- case opts of
                [x] | "-O" `isPrefixOf` x -> return $ drop 2 x
                [] -> return "a"   -- Default is to run all optimization levels
                _ -> printUsage
  let testNames = map f files
          where f ('i':'n':'p':'u':'t':'/':xs) = xs  -- strip out "input/" from the test name (if given)
                f xs = xs
  return (optLevel, testNames)

main :: IO ()
main = do
  (optLevel, testNames) <- processArgs
  xs <- getDirectoryContents "input"
  let filenames = case testNames of
                    [] -> filter (\x -> ".c" `isSuffixOf` x || ".cpp" `isSuffixOf` x || ".s" `isSuffixOf` x) xs
                    xs -> xs
  case optLevel of
    "a" -> mapM_ (\x -> runTests x filenames) ["0", "1", "2"]
    x -> runTests x filenames
  putStrLn ""
  putStrLn "*** All tests passed."
  exitSuccess

runTests :: String -> [String] -> IO ()
runTests optLevel filenames = do
  putStrLn ""
  putStrLn $ "Running tests at optimization level: " ++ optLevel
  mapM_ (runTest optLevel) filenames
